#include <FL/Fl.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/fl_draw.H>
#include <sys/types.h>
#include <vector>		// STL vector<>
#include <stdio.h>

// Fl_Table -- A table widget
//
//     1.00 erco@3dsite.com 12/03/02
//
// TODO:
//    o Auto scroll during dragged selection
//    o Keyboard navigation (up/down/left/right arrow)
//    o Row headings (only column headings supported currently)
//

class Fl_Table : public Fl_Group
{

public:
    // TABLE CONTEXT
    enum TableContext
    {
	CONTEXT_NONE		= 0,
	CONTEXT_NEWPAGE		= 1,	// before a page is redrawn
	CONTEXT_ROW_HEADER	= 2,	// in the row header
	CONTEXT_COL_HEADER	= 3,	// in the col header
	CONTEXT_CELL		= 4,	// in one of the cells
	CONTEXT_TABLE		= 5	// in the table
    };

private:
    int _rows, _cols;	 	// total rows/cols
    int rowh;			// row heights (all are same)

    char _row_header,		// row header enabled?
         _col_header,		// col header enabled?
	 _row_resize,		// row resizing enabled?
	 _col_resize;		// col resizing enabled?

    vector<int> colw;		// column widths (in pixels)

    Fl_Cursor last_cursor;	// what kind of mouse cursor?
    Fl_When _when;		// when to do callback (FL_WHEN_RELEASE, etc)

    // FLTK OVERRIDES AND CALLBACKS
    void resize(int X, int Y, int W, int H);	// fltk resize() override
    void draw(void);				// fltk draw() override
    static void scroll_cb(Fl_Widget*,void*);	// h/v scrollbar callback

protected:
    // PRIVATE: RESIZE FLAGS
    enum ResizeFlag
    {
        RESIZE_NONE      	= 0,
        RESIZE_COL_LEFT  	= 1,
        RESIZE_COL_RIGHT 	= 2
    };

    int tablew, tableh;				// table's virtual size (in pixels)
    int toprow, botrow, 			// four corners of viewable table
	leftcol, rightcol;
    Fl_Box       *tablebox;			// invisible box for table area
    Fl_Scrollbar *vscrollbar, 			// vertical scrollbar
                 *hscrollbar;			// horizontal scrollbar

    // FLTK
    int handle(int e);				// fltk handle() override

    // CLASS MAINTENANCE
    void table_window_size(int &X, int &Y, 	// return size of table window
    			   int &W, int &H);
    void table_resized();			// table resized; recalc
    void table_scrolled();			// table scrolled; recalc

    void get_bounds(TableContext context,	// return x/y/w/h bounds for context
                    int &X, int &Y, int &W, int &H);

    void change_cursor(Fl_Cursor newcursor);	// change mouse cursor to some other shape

    // find r/c given current x/y event
    Fl_Table::TableContext cursor2rowcol( int &R, int &C, ResizeFlag &resizeflag);

    int find_cell(TableContext context,		// find cell's x/y/w/h given r/c
    		 int R, int C,
    		 int &X, int &Y, 
		 int &W, int &H);

    int row_col_clamp(int &R, int &C);		// clamp r/c to known universe

    // Called to draw cells
    virtual void draw_cell(TableContext context, 
    			  int R=0, int C=0, 
			  int X=0, int Y=0, int W=0, int H=0)
    {
        // overridden by deriving class
    }

    // Event callback
    virtual void callback(TableContext context, int R, int C)
    {
        // overridden by deriving class
    }

public:
    Fl_Table(int X, int Y, int W, int H, const char *l=0);
    ~Fl_Table();

    int x() { return(Fl_Group::x() + 1); }	// +1: widget border compensation
    int y() { return(Fl_Group::y() + 1); }	// +1: widget border compensation
    int w() { return(Fl_Group::w() - 2); }	// -2: widget border compensation
    int h() { return(Fl_Group::h() - 2); }	// -2: widget border compensation

    void rows(int val);				// set/get number of rows
    inline int rows() { return(_rows); }

    void cols(int val);				// set/get number of columns
    inline int cols() { return(_cols); }

    void        row_resize(int flag) { _row_resize = flag; }	// enable row resizing
    inline int  row_resize()         { return(_row_resize); }

    void        col_resize(int flag) { _col_resize = flag; }	// enable col resizing
    inline int  col_resize()         { return(_col_resize); }

    void        row_header(int flag) { _row_header = flag; }	// enable row headers
    inline int  row_header()         { return(_row_header); }

    void        col_header(int flag) { _col_header = flag; }	// enable col headers
    inline int  col_header()         { return(_col_header); }

    void row_height(int row, int new_h);	// set/get row's height (currently all rows)
    int row_height(int row)
    {
        if ( row < 0 || row >= _rows ) return(-1);
        return(rowh);
    }

    void col_width(int col, int new_w);		// set/get a column's width
    int col_width(int col)
    {
        if ( col < 0 || col >= _cols ) return(-1);
        return(colw[col]);
    }

    int top_row() { return(toprow); }		// set/get top row
    void top_row(int row);

    void when(Fl_When val)			// Set 'when' table receives events
    {
        _when = val;
    }
};

// SCROLL DISPLAY SO 'row' IS AT THE TOP
void Fl_Table::top_row(int row)
{
    if ( row < 0 || row >= rows() ) return;
    double newtop = row * rowh;
    if ( newtop > vscrollbar->maximum() ) 
        { newtop = vscrollbar->maximum(); }
    vscrollbar->Fl_Slider::value(newtop);
    table_scrolled();
    redraw();
}

//////////////////////////////////////////////
//
// Fl_Table_Row
//
//    A class specializing in a table of rows.
//    Handles row-specific selection behavior.
//
//////////////////////////////////////////////

class Fl_Table_Row : public Fl_Table
{
private:
    vector<char> rowselect;			// selection flag for each row

    int handle(int event);

public:
    Fl_Table_Row(int X, int Y, int W, int H, const char *l=0) : Fl_Table(X,Y,W,H,l) { }
    ~Fl_Table_Row() { }
    
    void rows(int val);				// set number of rows
    int  rows() { return(Fl_Table::rows()); }	// get number of rows
    int row_selected(int row);			// is row selected? (0=no, 1=yes, -1=range err)
    int select_row(int row, int flag=1);	// select state for row: flag:0=off, 1=on, 2=toggle
						// returns: 0=no change, 1=changed, -1=range err
    void select_all_rows(int flag=1);		// all rows to a known state
};

// Is row selected?
int Fl_Table_Row::row_selected(int row)
{
    if ( row < 0 || row >= rows() ) return(-1);
    return(rowselect[row]);
}

// Change selection state for row
//
//     flag:
//        0 - clear selection
//        1 - set selection
//        2 - toggle selection
//
//     Returns:
//        0 - selection state did not change
//        1 - selection state changed
//       -1 - row out of range
//
int Fl_Table_Row::select_row(int row, int flag=1)
{
    if ( row < 0 || row >= rows() ) { return(-1); }

    int oldval = rowselect[row];

    if ( flag == 2 ) { rowselect[row] ^= 1; }
    else             { rowselect[row] = flag; }

    if ( rowselect[row] != oldval )		// select state changed?
    {
	if ( row >= toprow && row <= botrow )	// row visible?
	    { redraw(); }			// redraw
	return(1);
    }
    return(0);
}

// Select all rows to a known state
void Fl_Table_Row::select_all_rows(int flag = 1)
{
    for ( int row=0; row<(int)rowselect.size(); row++ )
	{ rowselect[row] = flag; }
    redraw();
}

// Set number of rows
void Fl_Table_Row::rows(int val)
{
    Fl_Table::rows(val);
    while ( val > (int)rowselect.size() ) { rowselect.push_back(0); }	// enlarge
    while ( val < (int)rowselect.size() ) { rowselect.pop_back(); }	// shrink
}

// #include "eventnames.h"		// debugging

// Handle events
int Fl_Table_Row::handle(int event)
{
    // fprintf(stderr, "** EVENT: %s: ", eventnames[event]);	// debugging

    // Let base class handle event
    int ret = Fl_Table::handle(event);
    if ( ret ) return(1);				// base class 'handled' it (eg. column resize)

    static int dragging_select = 0,			// dragging out a selection?
	       last_row = -1,
	       push_row = -1;				// FL_PUSH on this row (used by FL_RELEASE)

    static TableContext push_context = CONTEXT_NONE;	// FL_PUSH context

    int shiftstate = (Fl::event_state() & FL_CTRL) ? FL_CTRL :
                     (Fl::event_state() & FL_SHIFT) ? FL_SHIFT : 0;

    // Which row/column are we over?
    int R, C;  				// row/column being worked on
    ResizeFlag resizeflag;		// which resizing area are we over? (0=none)
    TableContext context = cursor2rowcol(R, C, resizeflag);

    switch ( event )
    {
	case FL_PUSH:

	    // Save for later when button released.
	    //     Do this for all mouse push buttons.
	    //
	    push_context = context;
	    push_row = R;

	    // Handle selection in table.
	    //     Select cell under cursor, and enable drag selection mode.
	    //
	    if ( context == CONTEXT_CELL && Fl::event_button() == 1 )
	    {
		// Ctrl key? Toggle selection state
		switch ( shiftstate )
		{
		    case FL_CTRL:
		        select_row(R, 2);	// toggle
			break;

		    case FL_SHIFT:
		    {
		        select_row(R, 1);
		        if ( last_row > -1 )
			{
			    int srow = R, erow = last_row;
			    if ( srow > erow ) 
				{ srow = last_row; erow = R; }
			    for ( int row = srow; row <= erow; row++ )
				{ select_row(row, 1); }
			}
			break;
		   }

		   default:
		        select_all_rows(0);	// clear all previous selections
			select_row(R, 1);
			break;
		}

		last_row = R;
		dragging_select = 1;
		ret = 1;			// FL_PUSH handled (ensures FL_DRAG will be received)
		// redraw();			// redraw() handled by select_row()
	    }

	    break;

        case FL_DRAG:

	    if ( dragging_select )
	    {
		switch ( shiftstate )
		{
		    case FL_CTRL:
			if ( R != last_row ) 		// toggle only if dragged to new row
			    { select_row(R, 2); } 	// 2=toggle
			break;

		    case FL_SHIFT:
		    default:
			select_row(R, 1);
			if ( last_row > -1 )
			{
			    int srow = R, erow = last_row;
			    if ( srow > erow ) 
				{ srow = last_row; erow = R; }
			    for ( int row = srow; row <= erow; row++ )
				{ select_row(row, 1); }
			}
			break;
		}
		ret = 1;			// drag handled
		last_row = R;
	    }
	    break;

	case FL_RELEASE:
	    if ( Fl::event_button() == 1 )
	    {
		dragging_select = 0;
		ret = 1;			// release handled
	    }
	    break;

	default:
	    break;
    }
    return(ret);
}

////////////////////////////////////////////////////////
//
// Fl_Table
//
//    A very low level base class; a generic row/column 
//    table of generic elements.
//
////////////////////////////////////////////////////////

// Ctor
Fl_Table::Fl_Table(int X, int Y, int W, int H, const char *l=0) : Fl_Group(X,Y,W,H,l)
{
    _rows       = 0;
    _cols       = 0;
    rowh        = 18;
    toprow      = 0;
    botrow      = 0;
    leftcol     = 0;
    rightcol    = 0;
    tablew      = 0;
    tableh      = 0;
    last_cursor = FL_CURSOR_DEFAULT;
    _when       = FL_WHEN_RELEASE;

    box(FL_THIN_DOWN_FRAME);

    tablebox = new Fl_Box(x(), y(), w(), h());
    tablebox->box(FL_NO_BOX);

    vscrollbar = new Fl_Scrollbar(x()+w()-17, y(), 17, h()-17);
    vscrollbar->type(FL_VERTICAL);
    vscrollbar->callback(scroll_cb, (void*)this);

    hscrollbar = new Fl_Scrollbar(x(), y()+h()-17, w(), 17);
    hscrollbar->type(FL_HORIZONTAL);
    hscrollbar->callback(scroll_cb, (void*)this);

    table_resized();

    end();
}

// Dtor
Fl_Table::~Fl_Table()
{
}

// Set height of a row.
//     Parent must call redraw().
//     Currently all rows are the same height.
//
void Fl_Table::row_height(int row, int new_h)
{
    if ( row < 0 || row >= _rows ) { return; }
    // TODO: Allow rows to be different sizes
    rowh = new_h;
}

// Set width of a row.
//     Parent must call redraw().
//
void Fl_Table::col_width(int col, int new_w)
{
    if ( col < 0 || col >= _cols ) { return; }
    colw[col] = new_w;
}

// Return row/col clamped to reality
int Fl_Table::row_col_clamp(int &R, int &C)
{
    int clamped = 0;
    if ( R < 0 )      { R = 0;       clamped = 1; }
    if ( R >= _rows ) { R = _rows-1; clamped = 1; }
    if ( C < 0 )      { C = 0;       clamped = 1; }
    if ( C >= _cols ) { C = _cols-1; clamped = 1; }
    return(clamped);
}

// Return bounding region for given context
void Fl_Table::get_bounds(TableContext context, int &X, int &Y, int &W, int &H)
{
    switch ( context )
    {
        case CONTEXT_COL_HEADER:

	    // Column header clipping.
	    X = tablebox->x();
	    Y = tablebox->y() - rowh;
	    W = tablebox->w();
	    H = rowh;
	    return;

	case CONTEXT_TABLE:

	    // Table window clipping.
	    //     If table window touches any Fl_Table widget edge, 
	    //     compensate for the 1 pixel border all around its inside.
	    //
	    X = tablebox->x();
	    Y = tablebox->y();
	    W = tablebox->w();
	    H = tablebox->h();
	    return;

	// TODO: Add other contexts..
	default:
	    fprintf(stderr, "get_bounds(): context %d unimplemented\n", context);
	    return;
    }
    //NOTREACHED
}

// Find row/col beneath cursor
//
//    Returns R/C and context.
//    Also returns resizeflag, if mouse is hovered over a resize boundary.
//
Fl_Table::TableContext Fl_Table::cursor2rowcol(int &R, int &C, ResizeFlag &resizeflag)
{
    R = C = 0;
    resizeflag = RESIZE_NONE;
    
    int X, Y, W, H;

    // Event inside a scrollbar?
    //    Then event is not in table context.
    //
    if ( ( vscrollbar->visible() && Fl::event_inside(vscrollbar) ) || 
         ( hscrollbar->visible() && Fl::event_inside(hscrollbar) ) )
    {
        return(CONTEXT_NONE);
    }
    
    // Column header?
    if ( col_header() )
    {
        get_bounds(CONTEXT_COL_HEADER, X, Y, W, H);

	// Inside a column heading?
	if ( Fl::event_inside(X, Y, W, H) )
	{
	    // Scan visible columns until found
	    for ( C = leftcol; C <= rightcol; C++ )
	    {
		find_cell(CONTEXT_COL_HEADER, 0, C, X, Y, W, H);
		if ( Fl::event_x() >= X && Fl::event_x() < (X+W) )
		{
		    // Found column?
		    //     If cursor over resize boundary, and resize enabled,
		    //     enable the appropriate resize flag.
		    //
		    if ( col_resize() )
		    {
		        if ( Fl::event_x() <= (X+3-0) ) { resizeflag = RESIZE_COL_LEFT; }
		        if ( Fl::event_x() >= (X+W-3) ) { resizeflag = RESIZE_COL_RIGHT; }
		    }
		    return(CONTEXT_COL_HEADER);
		}
	    }

	    // Must be in column header dead zone
	    return(CONTEXT_NONE);
	}
    }

    // Mouse somewhere in table?
    //     Scan visible r/c's until we find it.
    //
    for ( R = toprow; R <= botrow; R++ )
    {
	for ( C = leftcol; C <= rightcol; C++ )
	{
	    find_cell(CONTEXT_CELL, R, C, X, Y, W, H);
	    if ( Fl::event_inside(X, Y, W, H) )
	        { return(CONTEXT_CELL); }	// found it
	}
    }

    // Somewhere else in table
    return(CONTEXT_NONE);
}

// Find X/Y/W/H for cell at R/C
//     If R or C are out of range, returns -1 
//     with X/Y/W/H set to zero.
//
int Fl_Table::find_cell(TableContext context, int R, int C, int &X, int &Y, int &W, int &H)
{
    if ( row_col_clamp(R, C) )
	{ X=Y=W=H=0; return(-1); }

    for ( int c=X=0; c<C; c++ )
	{ X += colw[c]; }

    X = X - hscrollbar->value() + tablebox->x();
    Y = R * rowh - vscrollbar->value() + tablebox->y();
    W = colw[C];
    H = rowh;

    switch ( context )
    {
	case CONTEXT_COL_HEADER:
	    Y = tablebox->y() - rowh;		// force vertical position above tablebox
	    return(0);

	case CONTEXT_ROW_HEADER:
	    X = ( colw.size() == 0 ) ? 		// force horiz position left of tablebox
	        tablebox->x() : (tablebox->x() - colw[0]);
	    return(0);

	case CONTEXT_CELL:
	    return(0);

	// TODO -- HANDLE OTHER CONTEXTS
	default:
	    fprintf(stderr, "find_cell: unknown context %d\n", (int)context);
	    return(-1);
    }
    //NOTREACHED
}

// Return table window size
void Fl_Table::table_window_size(int &X, int &Y, int &W, int &H)
{
    X=x(); Y=y(); H=h(); W=w();

    // Trim window if scrollbars visible
    if ( vscrollbar->visible() ) { W -= 17; }
    if ( hscrollbar->visible() ) { H -= 17; }

    // Trim window if header enabled
    if ( col_header() )
    {
        Y += rowh;
	H -= rowh;
    }
}

// Recalculate internals after a scroll.
//
//    Call this if table has been scrolled or resized.
//    Does not handle redraw().
//    TODO: Assumes tablebox has already been recalculated.
//
void Fl_Table::table_scrolled()
{
    if ( _rows == 0 || _cols == 0 )
        return;

    // Top row
    int voff = vscrollbar->value();
    toprow = voff / rowh;
    if ( toprow >= _rows ) toprow = _rows - 1;

    // Bottom row
    voff = vscrollbar->value();
    botrow = ( voff + tablebox->h() ) / rowh;
    if ( botrow >= _rows ) botrow = _rows - 1;

    // Left column
    int x, hoff = hscrollbar->value();
    for ( leftcol=x=0; leftcol < _cols; leftcol++ )
    {
        x += colw[leftcol];
        if ( x >= hoff ) { x -= colw[leftcol]; break; }
    }
    leftcol = ( leftcol >= _cols ) ? (leftcol - 1) : leftcol;

    // Right column
    //    Work with data left over from leftcol calculation
    //
    hoff = hscrollbar->value() + w();
    for ( rightcol=leftcol; rightcol < _cols; rightcol++ )
    {
        x += colw[rightcol];
        if ( x >= hoff ) { break; }
    }
    rightcol = ( rightcol >= _cols ) ? (rightcol - 1) : rightcol;
}

// Table resized: recalc internal data
//    Call this whenever the window is resized.
//    Recalculates the scrollbar sizes.
//    Makes no assumptions about any pre-initialized data.
//
void Fl_Table::table_resized()
{
    // Height
    tableh = rowh * _rows;

    // Width
    for ( int col=tablew=0; col < (int)colw.size(); col++ )
        { tablew += colw[col]; }
    
    //
    // Make scroll bars disappear if window large enough
    //

    // First pass: can hide via window size?
    int hidev = ( tableh - h() < 0 ),
	hideh = ( tablew - w() < 0 );

    // Second pass: Check for interference
    if ( !hideh & hidev ) { hidev = ( ( tableh - h() + 17 ) < 0 ); } 
    if ( !hidev & hideh ) { hideh = ( ( tablew - w() + 17 ) < 0 ); }

    // Determine scrollbar visibility
    if ( hidev ) { vscrollbar->hide(); } else { vscrollbar->show(); }
    if ( hideh ) { hscrollbar->hide(); } else { hscrollbar->show(); }

    //
    // Get the table's physical clipping window.
    //
    //    Do this AFTER hide/showing scrollbars
    //    since if they're shown, they affect 
    //    the size of the table window.
    //
    int twx, twy, tww, twh;
    table_window_size(twx, twy, tww, twh);
    tablebox->resize(twx, twy, tww, twh);
    tablebox->redraw();

    // Recalc scrollbar sizes
    //    Clamp scrollbar value() after a resize.
    //    Resize scrollbars to enforce a constant trough width after a window resize.
    //
    {
        float vscrolltab = ( tableh == 0 || twh > tableh ) ? 1 : (float)twh / tableh;
        float hscrolltab = ( tablew == 0 || tww > tablew ) ? 1 : (float)tww / tablew;

	vscrollbar->bounds(0, tableh-twh);
	vscrollbar->precision(10);
	vscrollbar->slider_size(vscrolltab);
	vscrollbar->resize(x()+w()-17, y(), 17, h() - ((hideh)?0:17) );
	vscrollbar->Fl_Valuator::value(vscrollbar->clamp(vscrollbar->value()));	

	hscrollbar->bounds(0, tablew-tww);
	hscrollbar->precision(10);
        hscrollbar->slider_size(hscrolltab);
	hscrollbar->resize(x(), y()+h()-17, w() - ((hidev)?0:17), 17);
	hscrollbar->Fl_Valuator::value(hscrollbar->clamp(hscrollbar->value()));
    }

    // Tell FLTK child widgets were resized
    Fl_Group::init_sizes();

    // Recalc top/bot/left/right
    table_scrolled();

    // Redraw everything
    redraw();
}

// Someone moved a scrollbar
void Fl_Table::scroll_cb(Fl_Widget*w, void *data)
{
    Fl_Table *o = (Fl_Table*)data;
    o->table_scrolled();
    o->redraw();
}

// Set number of rows
void Fl_Table::rows(int val)
{
    _rows = val;
    table_resized();
}

// Set number of cols
void Fl_Table::cols(int val)
{
    _cols = val;
    {
	int default_w = ( colw.size() > 0 ) ? colw.back() : 80;
	while ( val > (int)colw.size() ) { colw.push_back(default_w); }	// enlarge
	while ( val < (int)colw.size() ) { colw.pop_back(); }		// shrink
    }

    table_resized();
}

// Change mouse cursor to different type
void Fl_Table::change_cursor(Fl_Cursor newcursor)
{
    if ( newcursor != last_cursor )
    {
	fl_cursor(newcursor, FL_BLACK, FL_WHITE);
	last_cursor = newcursor;
    }
}

// #include "eventnames.h"

// Handle FLTK events
int Fl_Table::handle(int event)
{
    // fprintf(stderr, "** EVENT: %s: ", eventnames[event]);

    static int resizing_col = -1,			// column being dragged
	       dragging_x = -1,				// starting x position for drag
	       push_row = -1,				// FL_PUSH on this row (used by FL_RELEASE)
	       push_col = -1;				// FL_PUSH on this col (used by FL_RELEASE)

    static TableContext push_context = CONTEXT_NONE;	// FL_PUSH context

    int ret = Fl_Group::handle(event);	// let FLTK group handle events first

    // Which row/column are we over?
    int R, C;  				// row/column being worked on
    ResizeFlag resizeflag;		// which resizing area are we over? (0=none)
    TableContext context = cursor2rowcol(R, C, resizeflag);

    switch ( event )
    {
	case FL_PUSH:

	    // Save these for later when button released.
	    //    Do this for all mouse push buttons.
	    //
	    push_context = context;
	    push_row = R;
	    push_col = C;

	    if ( context == CONTEXT_COL_HEADER && Fl::event_button() == 1 && resizeflag )
	    {
		// Start resize if left mouse clicked on column border.
		//
		//    "ret = 1" ensures we get drag events from now on.
		//    (C-1) is used if mouse is over the left hand side of the cell,
		//    so that we resize the next column over to the left.
		//
		resizing_col = ( resizeflag & RESIZE_COL_LEFT ) ? C-1 : C; 
		dragging_x = Fl::event_x(); 
		ret = 1;
	    }

	    break;

        case FL_DRAG:

	    // Dragging column?
	    //
	    //    Let user drag even /outside/ the row/col widget.
	    //    Don't allow column width smaller than 1.
	    //    Continue to show FL_CURSOR_WE at all times during drag.
	    //
	    if ( resizing_col > -1 )
	    {
		int offset = dragging_x - Fl::event_x();
		colw[resizing_col] -= offset;
		if ( colw[resizing_col] < 1 )
		    { colw[resizing_col] = 1; }
		dragging_x = Fl::event_x();
		table_resized();
		change_cursor(FL_CURSOR_WE);
		ret = 1;
	    }
	    break;

	case FL_RELEASE:
	    switch ( context )
	    {
		case CONTEXT_COL_HEADER:	// release on header
		case CONTEXT_CELL:		// release on a cell

		    // FL_RELEASE on same row/col we FL_PUSH'ed?
		    //    Then do the user's callback with context.
		    //
		    if ( _when & (FL_WHEN_RELEASE | FL_WHEN_RELEASE_ALWAYS ) &&	// when qualifies?
		         push_context == context && 				// same context?
		         push_row == R && 					// same row FL_PUSH'ed?
			 push_col == C && 					// same col FL_PUSH'ed?
			 resizing_col == -1 )					// not resizing a column
		    {
		        callback(context, R, C);	// don't use Fl_Table::callback(),
							// or parent class can't override
		    }
		    break;

		default:
		    break;
	    }

	    if ( Fl::event_button() == 1 )
	    {
		change_cursor(FL_CURSOR_DEFAULT);
		resizing_col = -1;
		ret = 1;
	    }
	    break;

	case FL_MOVE:
	    if ( context == CONTEXT_COL_HEADER && 	// in column header?
	         resizeflag )				// and resize enabled + near resize boundary?
		{ change_cursor(FL_CURSOR_WE); }	// show resize cursor
	    else
		{ change_cursor(FL_CURSOR_DEFAULT); }	// normal cursor
	    ret = 1;
	    break;

	default:
	    change_cursor(FL_CURSOR_DEFAULT);
	    break;
    }

    return(ret);
}

// Resize FLTK override
//     Handle resize events if user resizes parent window.
//
void Fl_Table::resize(int X, int Y, int W, int H)
{
    // Tell group to resize, and recalc our own widget as well
    Fl_Group::resize(X, Y, W, H);
    table_resized();
}

// Draw the entire Fl_Table
//    Override the draw() routine to draw the table.
//    Then tell the group to draw over us.
//
void Fl_Table::draw()
{
    int X,Y,W,H;

    draw_cell(CONTEXT_NEWPAGE);		// let user's drawing routine prepare for new page

    // Draw column headers
    if ( col_header() )
    {
	// Draw column header *above* 'tablebox'; compensate y clipping.
	fl_push_clip(tablebox->x(), tablebox->y()-rowh, tablebox->w(), rowh);

	for ( int c = leftcol; c <= rightcol; c++ )
	{
	    find_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	    draw_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	}
	fl_pop_clip();
    }

    // Draw all cells.
    //    This includes cells partially obscured off edges of table.
    //
    int cx, cy, cw, ch;
    get_bounds(CONTEXT_TABLE, cx, cy, cw, ch);
    fl_push_clip(cx, cy, cw, ch);
    {
	for ( int r = toprow; r <= botrow; r++ )
	{
	    for ( int c = leftcol; c <= rightcol; c++ )
	    {
		find_cell(CONTEXT_CELL, r, c, X, Y, W, H);
		draw_cell(CONTEXT_CELL, r, c, X, Y, W, H);
	    }
	}
    }
    fl_pop_clip();

    // Dead zone right of table visible?
    if ( cw > tablew )
    {
        // Fill in with a rectangle
        int X = cx + cw - ( cw - tablew ), Y = y(),
	    W = cw - tablew, H = h();
        fl_color(color());
	fl_rectf(X,Y,W,H);
    }

    // Dead zone below table visible?
    if ( ch > tableh )
    {
        // Fill in with a rectangle
        int X = cx, Y = y() + ch - ( ch - tableh ),
	    W = w(), H = ch - tableh;

	// TODO
	//     Shouldn't have to do this. We probably need another tablebox-like
	//     object that has the dimensions of the ENTIRE table, including headers.
	//
	if ( col_header() ) { Y += rowh; } 
        fl_color(color());
	fl_rectf(X,Y,W,H);
    }

    // Dead zone between scrollbars?
    if ( vscrollbar->visible() && hscrollbar->visible() )
    {
        fl_color(color());
	fl_rectf(x()+w()-17, y()+h()-17, 17, 17);
    }

    // Draw group over the above
    //
    //     Group's background is transparent (a frame)
    //     so this really just draws the thin frame around the table.
    //
    Fl_Group::draw();
}

///////////////////////
//
// STANDALONE TEST CODE
//
///////////////////////

#define STANDALONE
#ifdef STANDALONE

#include <FL/Fl_Window.H>
#include <string.h>
#include <iostream.h>
#include <algorithm>		// STL sort

// Row class; contains columns
class Row
{
public:
    vector<char*> words;
};

// Sort class to handle sorting strings by column
class SortColumn
{
    int _col, _reverse;
public:
    SortColumn(int col, int reverse)
    {
	_col = col;
	_reverse = reverse;
    }
    bool operator()(const Row &a, const Row &b)
    {
	if ( _col >= (int)a.words.size() ) return(1);
	if ( _col >= (int)b.words.size() ) return(0);
	if ( _reverse )
	    return(strcmp(a.words[_col], b.words[_col]) > 0);
	else
	    return(strcmp(a.words[_col], b.words[_col]) < 0);
    }
};

// Demonstrate how to use Fl_Table_Row
class MyTable : public Fl_Table_Row
{
private:
    vector<Row> _rowdata;	// data in each row
    int _maxcols;		// max # columns in all rows

protected:
    void draw_cell(TableContext context, int R=0, int C=0, 	// table cell drawing
    		   int X=0, int Y=0, int W=0, int H=0);
    void sort_column(int col, int reverse=0);			// sort table by a column
    void callback(TableContext context, int R, int C);		// callback for table events

public:
    MyTable(int x, int y, int w, int h, const char *l=0) : Fl_Table_Row(x,y,w,h,l)
	{ _maxcols = 0; }
    ~MyTable() { }

    void load_command(const char *cmd);		// load the output of a command into table
    void autowidth(int pad);			// automatically set column widths to data
};

// Sort a column up or down
void MyTable::sort_column(int col, int reverse)
{
    sort(_rowdata.begin(), _rowdata.end(), SortColumn(col, reverse));
    redraw();
}

// Handle drawing all cells in table
void MyTable::draw_cell(TableContext context, int R=0, int C=0, 
		        int X=0, int Y=0, int W=0, int H=0)
{
    char *s = "";

    if ( C < (int)_rowdata[R].words.size() )
	s = _rowdata[R].words[C];

    switch ( context )
    {
	case CONTEXT_NEWPAGE:
	    fl_font(FL_COURIER, 16);
	    return;

	case CONTEXT_ROW_HEADER:
	    fl_color(FL_RED);
	    fl_rectf(X, Y, W, H);
	    return;

	case CONTEXT_COL_HEADER:
	    fl_push_clip(X, Y, W, H);
	    {
		static char *head[] = 
		    { "Perms", "#L", "Owner", "Group", "Size", "Date", "", "", "Filename" };

		// HEADER BG
		fl_color(color());
		fl_rectf(X+1, Y+1, W-2, H-2);

		// HEADER TEXT
		if ( C < 9 )
		{
		    fl_color(FL_BLACK);
		    fl_draw(head[C], X, Y, W, H, FL_ALIGN_LEFT, 0, 0);
		}

		// HEADER 3-D
		fl_color(FL_LIGHT3);
		fl_line(X, Y, X+W-1, Y);		// horiz highlight
		fl_line(X, Y, X, Y+H-1);		// vert highlight

		// HEADER BORDER
		fl_color(FL_DARK3);
		fl_line(X, Y+H-1, X+W, Y+H-1);		// horiz dark
		fl_line(X+W-1, Y, X+W-1, Y+H-1);	// vert dark
	    }
	    fl_pop_clip();
	    return;

	case CONTEXT_CELL:
	{
	    // DRAW CELLS WITH ROW AND COLUMN DIVIDERS
	    fl_color( row_selected(R) ? selection_color() : FL_WHITE );
	    fl_rectf(X, Y, W, H);

	    fl_color(color());
	    fl_line(X, Y+H-1, X+W-1, Y+H-1);		// horiz line
	    fl_line(X+W-1, Y, X+W-1, Y+H);		// vert line

	    // DRAW THE TEXT
	    //    Clip text within the row/column dividers
	    //
	    fl_push_clip(X, Y, W, H);
	    {
		fl_color(FL_BLACK);
		fl_draw(s, X, Y, W, H, FL_ALIGN_LEFT, 0, 0);
	    }
	    fl_pop_clip();
	    return;
	}
    }
}

// Automatically set column widths to widest data in each column
void MyTable::autowidth(int pad)
{
    fl_font(FL_COURIER, 16);

    // Initialize all column widths to lowest value
    for ( int c=0; c<cols(); c++ )
	{ col_width(c, pad); }

    for ( int r=0; r<(int)_rowdata.size(); r++ )
    {
	int w, h;
	for ( int c=0; c<(int)_rowdata[r].words.size(); c++ )
	{
	    fl_measure(_rowdata[r].words[c], w, h, 0);
	    if ( (w + pad) > col_width(c))
		{ col_width(c, w + pad); }
	}
    }
    table_resized();
    redraw();
}

// Load table with output of 'cmd'
void MyTable::load_command(const char *cmd)
{
    char s[512];
    FILE *fp = popen(cmd, "r");

    _maxcols = 0;

    for ( int r=0; fgets(s, sizeof(s)-1, fp); r++ )
    {
	if ( strlen(s) > 1 )
	    s[strlen(s)-1] = 0;

	// new row
	Row newrow;
	_rowdata.push_back(newrow);

	char *ss = s, *word = s;
	while ( *ss )
	{
	    // Found white space? 
	    if ( strchr(" \t", *ss) )
	    {
		// Make previous data a word, add to column
		{
		    char tmp = *ss;
		    *ss = 0;
		    _rowdata[r].words.push_back(strdup(word));
		    *ss = tmp;
		}

		// Skip over all white
		while ( strchr(" \t", *ss) )
		    { ++ss; }

		// Next word starts at first non-white
		word = ss;
	    }
	    ss++;
	}

	// last word
	_rowdata[r].words.push_back(strdup(word));

	// keep track of most # columns
	if ( (int)_rowdata[r].words.size() > _maxcols )
	    { _maxcols = (int)_rowdata[r].words.size(); }
    }

    // Initialize table rows/cols based on data loaded
    rows((int)_rowdata.size());
    cols(_maxcols);

    // Auto-calculate widths, with 20 pixel padding
    autowidth(20);
}

// Callback whenever someone clicks on different parts of the table
void MyTable::callback(TableContext context, int R, int C)
{
    switch ( context )
    {
	case CONTEXT_COL_HEADER:		// someone clicked on column header
	{
	    static int sort_reverse = 0,
		       sort_lastcol = -1;

	    if ( sort_lastcol == C )
		{ sort_reverse ^= 1; } 		// Click same column? Toggle sort
	    else
		{ sort_reverse = 0; } 		// Click diff column? Up sort 

	    sort_column(C, sort_reverse);
	    sort_lastcol = C;
	    break;
	}
    }
}

int main()
{
    Fl_Window win(900,500);
    MyTable table(20, 20, win.w()-40, win.h()-40);
    table.selection_color(FL_YELLOW);
    table.col_header(1);
    table.col_resize(1);
    table.when(FL_WHEN_RELEASE);		// handle table events on release
#if defined(_WIN32)
    table.load_command("dir \\winnt");		// load table with a directory listing
#else
    table.load_command("ls -la /var/tmp");	// load table with a directory listing
#endif
    win.end();
    win.resizable(table);
    win.show();
    return(Fl::run());
}
#endif /*STANDALONE*/
